@@ -1,22 +1,24 @@
 import pysam
 import numpy as np
-import os, sys
+import os
+import sys
 import pandas as pd
 from multiprocessing import Pool
 import functools
 from scipy import stats
-import pysam
 import tqdm
 import logging
 import warnings
 from bam_filter.utils import is_debug, calc_chunksize, initializer
 from bam_filter.entropy import entropy, norm_entropy, gini_coeff, norm_gini_coeff
-from collections import Counter, defaultdict
+from collections import defaultdict
 import pyranges as pr
 
 # import cProfile as profile
 # import pstats
 
+# import pstats
+
 import matplotlib.pyplot as plt
 
 log = logging.getLogger("my_logger")
@@ -25,6 +27,17 @@ log = logging.getLogger("my_logger")
 sys.setrecursionlimit(10**6)
 
 
+# def get_alns(params):
+#     bam, reference = params
+#     samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
+#     alns = []
+#     for aln in samfile.fetch(
+#         reference=reference, multiple_iterators=False, until_eof=True
+#     ):
+#         alns.append(aln.to_string())
+#     return alns
+
+
 def get_tad(cov, trim_min=10, trim_max=90):
     """
     Get the TAD of a coverage
@@ -33,10 +46,10 @@ def get_tad(cov, trim_min=10, trim_max=90):
         (cov >= np.percentile(cov, trim_min)) & (cov <= np.percentile(cov, trim_max))
     ]
 
-    if sum(cov) == 0:
+    if np.sum(cov) == 0:
         return 0, 0
     else:
-        return sum(cov) / len(cov), len(cov)
+        return np.sum(cov) / len(cov), len(cov)
 
 
 # Function to calculate evenness of coverage
@@ -63,7 +76,7 @@ def coverage_evenness(coverage):
         covEvenness = 1.0
     else:
         if C > 0:
-            covEvenness = 1.0 - (len(D2) - sum(D2) / C) / len(coverage)
+            covEvenness = 1.0 - (len(D2) - np.sum(D2) / C) / len(coverage)
         else:
             covEvenness = 0.0
 
@@ -117,6 +130,7 @@ def get_bam_stats(
     plot=False,
     plots_dir="coverage-plots",
     read_length_freqs=False,
+    threads=1,
 ):
     """
     Worker function per chromosome
@@ -127,7 +141,8 @@ def get_bam_stats(
     # prof.enable()
     bam, references = params
     results = []
-    samfile = pysam.AlignmentFile(bam, "rb")
+    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
+
     read_hits = defaultdict(int)
     for reference in references:
         edit_distances = []
@@ -157,7 +172,7 @@ def get_bam_stats(
         cov_np = np.zeros(samfile.get_reference_length(reference), dtype=int)
 
         for aln in samfile.fetch(
-            contig=reference, multiple_iterators=False, until_eof=False
+            contig=reference, multiple_iterators=False, until_eof=True
         ):
             ani_read = (1 - ((aln.get_tag("NM") / aln.infer_query_length()))) * 100
             if ani_read >= min_read_ani:
@@ -226,7 +241,8 @@ def get_bam_stats(
         ranges = create_pyranges(reference, starts, ends, strands)
         if ranges.df.shape[0] == 0:
             log.debug(f"No alignments found for {reference}")
-            return None, None, read_hits
+            results.append(None)
+            continue
         ranges_raw = ranges.merge(strand=False)
         ranges = ranges_raw.lengths().to_list()
 
@@ -237,8 +253,8 @@ def get_bam_stats(
         cov_sd = np.std(cov_pos, ddof=1)
         cov_var = np.var(cov_pos, ddof=1)
         # get average coverage
-        mean_coverage = sum(cov_pos) / (reference_length - (2 * trim_ends))
-        mean_coverage_covered = sum(cov_pos) / bases_covered
+        mean_coverage = np.sum(cov_pos) / (reference_length - (2 * trim_ends))
+        mean_coverage_covered = np.sum(cov_pos) / bases_covered
 
         breadth = bases_covered / (reference_length - (2 * trim_ends))
         exp_breadth = 1 - np.exp(-mean_coverage)
@@ -392,6 +408,12 @@ def get_bam_stats(
         )
         results.append(data)
     samfile.close()
+    # prof.disable()
+    # # print profiling output
+    # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
+    # stats.print_stats(5)  # top 10 rows
+    # exit()
+    # results = [x[0] for x in results if x[0] is not None]
     results = list(filter(None, results))
     data_df = pd.DataFrame([x.to_summary() for x in results])
     # read_hits = (
@@ -400,14 +422,14 @@ def get_bam_stats(
     #     .reset_index()
     # )
     if read_length_freqs:
-        read_lens = [x.get_read_length_freqs for x in results]
-        return data_df, read_lens, read_hits
+        read_lens = [x.get_read_length_freqs() for x in results]
+        return (data_df, read_lens, read_hits)
     else:
-        return data_df, None, read_hits
-    # prof.disable()
-    # # print profiling output
-    # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
-    # stats.print_stats(5)  # top 10 rows
+        return (
+            data_df,
+            None,
+            read_hits,
+        )
 
 
 class BamAlignment:
@@ -602,7 +624,7 @@ class BamAlignment:
         frags = {}
         lengths = pd.Series(self.read_length)
         lengths = lengths.value_counts().sort_index()
-        freqs = list(lengths / sum(lengths))
+        freqs = list(lengths / np.sum(lengths))
         frags[self.reference] = {"length": list(lengths.index), "freq": freqs}
         return frags
 
@@ -629,9 +651,9 @@ def process_bam(
     to extract from a bam file two definitions of the edit distances to the reference genome scaled by read length
     Returned in a pandas DataFrame
     """
-    logging.info(f"Loading BAM file")
+    logging.info("Loading BAM file")
     save = pysam.set_verbosity(0)
-    samfile = pysam.AlignmentFile(bam, "rb")
+    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
 
     references = samfile.references
 
@@ -649,7 +671,7 @@ def process_bam(
         # check if the dataframe contains all the References in the BAM file
         if not set(references).issubset(set(ref_lengths.index)):
             logging.error(
-                f"The BAM file contains references not found in the reference lengths file"
+                "The BAM file contains references not found in the reference lengths file"
             )
             sys.exit(1)
         max_chr_length = np.max(ref_lengths["length"].tolist())
@@ -660,7 +682,7 @@ def process_bam(
         sorted_bam = bam.replace(".bam", ".bf-sorted.bam")
         pysam.sort("-@", str(threads), "-m", str(sort_memory), "-o", sorted_bam, bam)
         bam = sorted_bam
-        samfile = pysam.AlignmentFile(bam, "rb")
+        samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
 
     if not samfile.has_index():
         logging.info("BAM index not found. Indexing...")
@@ -676,7 +698,7 @@ def process_bam(
 
         logging.info("Reloading BAM file")
         samfile = pysam.AlignmentFile(
-            bam, "rb"
+            bam, "rb", threads=threads
         )  # Need to reload the samfile after creating index
 
     total_refs = samfile.nreferences
@@ -739,6 +761,7 @@ def process_bam(
                         plot=plot,
                         plots_dir=plots_dir,
                         read_length_freqs=read_length_freqs,
+                        threads=threads,
                     ),
                     params,
                 )
@@ -760,9 +783,12 @@ def process_bam(
                             min_read_ani=min_read_ani,
                             trim_ends=0,
                             trim_min=trim_min,
+                            trim_max=trim_max,
                             scale=scale,
                             plot=plot,
                             plots_dir=plots_dir,
+                            read_length_freqs=read_length_freqs,
+                            threads=threads,
                         ),
                         params,
                         chunksize=1,
@@ -770,7 +796,7 @@ def process_bam(
                     total=len(ref_chunks),
                     leave=False,
                     ncols=80,
-                    desc=f"References processed",
+                    desc="References processed",
                 )
             )
 
@@ -778,7 +804,7 @@ def process_bam(
             p.join()
 
     except KeyboardInterrupt:
-        logging.info(f"User canceled the operation. Terminating jobs.")
+        logging.info("User canceled the operation. Terminating jobs.")
         p.terminate()
         p.join()
         sys.exit(0)
@@ -793,6 +819,7 @@ def filter_reference_BAM(
     out_files,
     sort_memory,
     only_stats_filtered,
+    transform_cov_evenness=False,
     sort_by_name=False,
 ):
     """Filter BAM based on certain conditions
@@ -806,8 +833,16 @@ def filter_reference_BAM(
     logging.info("Filtering stats...")
     if "min_norm_entropy" in filter_conditions and "min_norm_gini" in filter_conditions:
         logging.info(
-            f"min_read_count >= {filter_conditions['min_read_count']} & min_read_length >= {filter_conditions['min_read_length']} & min_avg_read_ani >= {filter_conditions['min_avg_read_ani']} & min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} &  min_breadth >= {filter_conditions['min_breadth']} & min_coverage_evenness >= {filter_conditions['min_coverage_evenness']} & min_norm_entropy >= {filter_conditions['min_norm_entropy']} & min_norm_gini <= {filter_conditions['min_norm_gini']}"
+            f"::: min_read_count >= {filter_conditions['min_read_count']} & min_read_length >= {filter_conditions['min_read_length']} & min_avg_read_ani >= {filter_conditions['min_avg_read_ani']} & min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} & min_breadth >= {filter_conditions['min_breadth']} & min_coverage_evenness >= {filter_conditions['min_coverage_evenness']} & min_coverage_mean >= {filter_conditions['min_coverage_mean']} & min_norm_entropy >= {filter_conditions['min_norm_entropy']} & min_norm_gini <= {filter_conditions['min_norm_gini']}"
         )
+        # We transform the coverage_evenenness to 1.0 where the coverage is smaller than 1
+        if transform_cov_evenness is True:
+            df["cov_evenness_tmp"] = df["cov_evenness"]
+            df["cov_evenness_tmp"] = np.where(
+                np.rint(df.coverage_mean) < 1.0, 1.0, df.cov_evenness_tmp
+            )
+        else:
+            df["cov_evenness_tmp"] = df["cov_evenness"]
         df_filtered = df.loc[
             (df["n_reads"] >= filter_conditions["min_read_count"])
             & (df["read_length_mean"] >= filter_conditions["min_read_length"])
@@ -817,14 +852,23 @@ def filter_reference_BAM(
                 >= filter_conditions["min_expected_breadth_ratio"]
             )
             & (df["breadth"] >= filter_conditions["min_breadth"])
-            & (df["cov_evenness"] >= filter_conditions["min_coverage_evenness"])
+            & (df["cov_evenness_tmp"] >= filter_conditions["min_coverage_evenness"])
+            & (df["coverage_mean"] >= filter_conditions["min_coverage_mean"])
             & (df["norm_entropy"] >= filter_conditions["min_norm_entropy"])
             & (df["norm_gini"] <= filter_conditions["min_norm_gini"])
         ]
     else:
         logging.info(
-            f"min_read_count >= {filter_conditions['min_read_count']} & min_read_length >= {filter_conditions['min_read_length']} & min_avg_read_ani >= {filter_conditions['min_avg_read_ani']} & min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} &  min_breadth >= {filter_conditions['min_breadth']} & min_coverage_evenness >= {filter_conditions['min_coverage_evenness']}"
+            f"::: min_read_count >= {filter_conditions['min_read_count']} & min_read_length >= {filter_conditions['min_read_length']} & min_avg_read_ani >= {filter_conditions['min_avg_read_ani']} & min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} & min_breadth >= {filter_conditions['min_breadth']} & min_coverage_evenness >= {filter_conditions['min_coverage_evenness']} & min_coverage_mean >= {filter_conditions['min_coverage_mean']}"
         )
+        if transform_cov_evenness is True:
+            df["cov_evenness_tmp"] = df["cov_evenness"]
+            df["cov_evenness_tmp"] = np.where(
+                np.rint(df.coverage_mean) < 1.0, 1.0, df.cov_evenness_tmp
+            )
+        else:
+            df["cov_evenness_tmp"] = df["cov_evenness"]
+
         df_filtered = df.loc[
             (df["n_reads"] >= filter_conditions["min_read_count"])
             & (df["read_length_mean"] >= filter_conditions["min_read_length"])
@@ -834,10 +878,17 @@ def filter_reference_BAM(
                 >= filter_conditions["min_expected_breadth_ratio"]
             )
             & (df["breadth"] >= filter_conditions["min_breadth"])
-            & (df["cov_evenness"] >= filter_conditions["min_coverage_evenness"])
+            & (df["cov_evenness_tmp"] >= filter_conditions["min_coverage_evenness"])
+            & (df["coverage_mean"] >= filter_conditions["min_coverage_mean"])
         ]
+
+    del df_filtered["cov_evenness_tmp"]
+    # prof = profile.Profile()
+    # prof.enable()
     if len(df_filtered.index) > 0:
-        logging.info(f"Saving filtered stats...")
+        # prof = profile.Profile()
+        # prof.enable()
+        logging.info("Saving filtered stats...")
         df_filtered.to_csv(
             out_files["stats_filtered"], sep="\t", index=False, compression="gzip"
         )
@@ -846,24 +897,26 @@ def filter_reference_BAM(
         else:
             logging.info("Writing filtered BAM file... (be patient)")
             refs_dict = dict(
-                zip(df_filtered["reference"], df_filtered["reference_length"])
+                zip(df_filtered["reference"], df_filtered["bam_reference_length"])
             )
             (ref_names, ref_lengths) = zip(*refs_dict.items())
 
-            out_bam_file = pysam.Samfile(
+            refs_idx = {x: i for i, x in enumerate(ref_names)}
+
+            out_bam_file = pysam.AlignmentFile(
                 out_files["bam_filtered_tmp"],
                 "wb",
                 referencenames=list(ref_names),
                 referencelengths=list(ref_lengths),
                 threads=threads,
             )
-            header = pysam.AlignmentHeader.from_references(
-                list(ref_names), list(ref_lengths)
-            )
-            references = df_filtered["reference"].values
 
-            logging.info("Filtering BAM file...")
-            samfile = pysam.AlignmentFile(bam, "rb")
+            samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
+            references = [x for x in samfile.references if x in refs_idx.keys()]
+
+            logging.info(
+                f"::: Filtering {len(references):,} references sequentially..."
+            )
             for reference in tqdm.tqdm(
                 references,
                 total=len(references),
@@ -874,10 +927,13 @@ def filter_reference_BAM(
                 for aln in samfile.fetch(
                     reference=reference, multiple_iterators=False, until_eof=True
                 ):
-                    out_bam_file.write(
-                        pysam.AlignedSegment.fromstring(aln.to_string(), header=header)
-                    )
+                    aln.reference_id = refs_idx[aln.reference_name]
+                    out_bam_file.write(aln)
             out_bam_file.close()
+            # prof.disable()
+            # # print profiling output
+            # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
+            # stats.print_stats(5)  # top 10 rows
             if sort_by_name:
                 logging.info("Sorting BAM file by read name...")
                 pysam.sort(
@@ -901,8 +957,11 @@ def filter_reference_BAM(
                     out_files["bam_filtered_tmp"],
                 )
 
+                logging.info("BAM index not found. Indexing...")
                 save = pysam.set_verbosity(0)
-                samfile = pysam.AlignmentFile(out_files["bam_filtered"], "rb")
+                samfile = pysam.AlignmentFile(
+                    out_files["bam_filtered"], "rb", threads=threads
+                )
                 chr_lengths = []
                 for chrom in samfile.references:
                     chr_lengths.append(samfile.get_reference_length(chrom))
@@ -910,7 +969,6 @@ def filter_reference_BAM(
                 pysam.set_verbosity(save)
                 samfile.close()
 
-                logging.info("BAM index not found. Indexing...")
                 if max_chr_length > 536870912:
                     logging.info("A reference is longer than 2^29, indexing with csi")
                     pysam.index(
@@ -929,12 +987,3 @@ def filter_reference_BAM(
             os.remove(out_files["bam_filtered_tmp"])
     else:
         logging.info("No references meet the filter conditions. Skipping...")
-
-
-def get_alns(params):
-    bam, reference = params
-    samfile = pysam.AlignmentFile(bam, "rb")
-    alns = []
-    for aln in samfile.fetch(reference=reference, multiple_iterators=False):
-        alns.append(aln.to_string())
-    return alns
